@@ -2,6 +2,7 @@ import numpy as np
 import numbers
 from copy import deepcopy
 import json
+import warnings
 
 
 def convert_numpy(obj):
@@ -245,6 +246,19 @@ class DecisionTreeAdversary:
         self.kind = kind
         self.one_adversarial_class = one_adversarial_class
 
+        if hasattr(decision_tree, "n_features_in_"):
+            if attack_model is None:
+                attack_model = ["" for _ in range(decision_tree.n_features_in_)]
+                warnings.warn(
+                    "Attack model not specified, assuming all features are not perturbable."
+                )
+
+            if is_numeric is None:
+                is_numeric = [True for _ in range(decision_tree.n_features_in_)]
+
+            if n_categories is None:
+                n_categories = [None for _ in range(decision_tree.n_features_in_)]
+
         if is_numeric is not None:
             self.is_numeric = is_numeric
 
@@ -290,7 +304,7 @@ class DecisionTreeAdversary:
             self.__calculate_leaves_sklearn()
 
     def __calculate_leaves_ours(self):
-        inf = 10.0 ** 10
+        inf = 10.0**10
         conditions = []
         for numeric in self.is_numeric:
             if numeric:
@@ -346,7 +360,7 @@ class DecisionTreeAdversary:
             return left_leaves + right_leaves
 
     def __calculate_leaves_sklearn(self):
-        inf = 10.0 ** 10
+        inf = 10.0**10
         conditions = []
         for numeric in self.is_numeric:
             if numeric:
@@ -420,7 +434,7 @@ class DecisionTreeAdversary:
             return left_leaves + right_leaves
 
     def __calculate_leaves_treant(self):
-        inf = 10.0 ** 10
+        inf = 10.0**10
         conditions = []
         self.is_numeric = self.decision_tree.numerical_idx
         for numeric in self.is_numeric:
