@@ -1,10 +1,10 @@
 import json
 import numbers
-import time
 import numpy as np
 from collections import defaultdict
 from itertools import product
 from numba import jit
+from numpy.lib.function_base import iterable
 from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
 from sklearn.utils.multiclass import type_of_target
 from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
@@ -40,7 +40,7 @@ class Node:
 
         return self.value
 
-    def pretty_print(self, depth=0):
+    def pretty_print(self, _, depth=0):
         indentation = depth * "  "
         if isinstance(self.value, np.ndarray):
             return f"{indentation}return [{self.value[0]:.3f}, {self.value[1]:.3f}]"
@@ -92,12 +92,18 @@ class NumericalNode(Node):
         else:
             return self.right_child.predict(sample)
 
-    def pretty_print(self, depth=0):
+    def pretty_print(self, feature_names, depth=0):
         indentation = depth * "  "
-        return f"""{indentation}if x{self.feature} <= {self.threshold}:
-{self.left_child.pretty_print(depth + 1)}
+
+        if feature_names:
+            feature_name = feature_names[self.feature]
+        else:
+            feature_name = f"x[{self.feature}]"
+
+        return f"""{indentation}if {feature_name} <= {self.threshold}:
+{self.left_child.pretty_print(feature_names, depth + 1)}
 {indentation}else:
-{self.right_child.pretty_print(depth + 1)}"""
+{self.right_child.pretty_print(feature_names, depth + 1)}"""
 
     def to_json(self):
         return {
@@ -159,24 +165,149 @@ class NumericalNode(Node):
             return self
 
 
-def _attack_model_to_tuples(attack_model):
-    new_attack_model = []
-    for attack_mode in attack_model:
-        if attack_mode == "":
-            new_attack_model.append((0, 0))
-        elif attack_mode == ">":
-            new_attack_model.append((0, 10e9))
-        elif attack_mode == "<":
-            new_attack_model.append((10e9, 0))
-        elif attack_mode == "<>":
-            new_attack_model.append((10e9, 10e9))
-        elif isinstance(attack_mode, numbers.Number):
-            new_attack_model.append((attack_mode, attack_mode))
-        elif isinstance(attack_mode, tuple) and len(attack_mode) == 2:
-            new_attack_model.append(attack_mode)
+def node_tree_to_arrays(node: Node):
+    xgboost_json, n_nodes = node.to_xgboost_json(0, 0)
+    n_nodes += 1
+
+    left_ids = np.empty(n_nodes, dtype=np.int32)
+    right_ids = np.empty(n_nodes, dtype=np.int32)
+    features = np.empty(n_nodes, dtype=np.int32)
+    thresholds = np.empty(n_nodes, dtype=np.float32)
+    values = np.empty(n_nodes, dtype=np.float32)
+
+    def _recurse(json_node):
+        node_id = json_node["nodeid"]
+
+        if "leaf" in json_node:
+            left_ids[node_id] = _TREE_LEAF
+            right_ids[node_id] = _TREE_LEAF
+            features[node_id] = _TREE_LEAF
+            thresholds[node_id] = _TREE_LEAF
+            values[node_id] = json_node["leaf"]
         else:
-            raise Exception("Unknown attack model spec:", attack_mode)
-    return new_attack_model
+            left_ids[node_id] = json_node["yes"]
+            right_ids[node_id] = json_node["no"]
+            features[node_id] = json_node["split"]
+            thresholds[node_id] = json_node["split_condition"]
+            values[node_id] = _TREE_UNDEFINED
+
+            _recurse(json_node["children"][0])
+            _recurse(json_node["children"][1])
+
+    _recurse(xgboost_json)
+
+    return left_ids, right_ids, features, thresholds, values
+
+
+@jit(nopython=True, nogil=NOGIL)
+def _predict_compiled(X, left_ids, right_ids, features, thresholds, values):
+    n_samples = X.shape[0]
+
+    # Initialize the output to -1
+    y_pred = np.empty(n_samples, dtype=np.float32)
+
+    # Iterate over the samples
+    for i, sample in enumerate(X):
+        # Initialize the current node to the root node
+        node_id = 0
+
+        # Iterate over the nodes until we reach a leaf
+        while True:
+            # Get the feature and threshold of the current node
+            feature = features[node_id]
+            threshold = thresholds[node_id]
+
+            # If the feature is -1, we have reached the leaf node
+            if feature == _TREE_LEAF:
+                break
+
+            # If the sample is lower or equal to the threshold, follow the left child
+            if sample[feature] <= threshold:
+                node_id = left_ids[node_id]
+            else:
+                node_id = right_ids[node_id]
+
+        # Store the prediction of the leaf node
+        y_pred[i] = values[node_id]
+
+    return y_pred
+
+
+class CompiledTree:
+    def __init__(self, node):
+        (
+            self.left_ids,
+            self.right_ids,
+            self.features,
+            self.thresholds,
+            self.values,
+        ) = node_tree_to_arrays(node)
+
+    def predict_classification(self, X):
+        pred_values = _predict_compiled(
+            X,
+            self.left_ids,
+            self.right_ids,
+            self.features,
+            self.thresholds,
+            self.values,
+        )
+        return (pred_values > 0.0).astype(np.int32)
+
+    def predict_classification_proba(self, X):
+        pred_values = _predict_compiled(
+            X,
+            self.left_ids,
+            self.right_ids,
+            self.features,
+            self.thresholds,
+            self.values,
+        )
+
+        # Rescale [-1, 1] values to probabilities in range [0, 1]
+        pred_values += 1
+        pred_values *= 0.5
+
+        return np.vstack([1 - pred_values, pred_values]).T
+
+    def predict_regression(self, X):
+        return _predict_compiled(
+            X,
+            self.left_ids,
+            self.right_ids,
+            self.features,
+            self.thresholds,
+            self.values,
+        )
+
+
+def _attack_model_to_tuples(attack_model, n_features):
+    if isinstance(attack_model, numbers.Number):
+        return [(attack_model, attack_model) for _ in range(n_features)]
+    elif iterable(attack_model):
+        new_attack_model = []
+        for attack_mode in attack_model:
+            if attack_mode == "":
+                new_attack_model.append((0, 0))
+            elif attack_mode == ">":
+                new_attack_model.append((0, 10e9))
+            elif attack_mode == "<":
+                new_attack_model.append((10e9, 0))
+            elif attack_mode == "<>":
+                new_attack_model.append((10e9, 10e9))
+            elif isinstance(attack_mode, numbers.Number):
+                new_attack_model.append((attack_mode, attack_mode))
+            elif isinstance(attack_mode, tuple) and len(attack_mode) == 2:
+                new_attack_model.append(attack_mode)
+            else:
+                raise Exception("Unknown attack model spec:", attack_mode)
+        return new_attack_model
+    else:
+        raise Exception(
+            "Unknown attack model spec, needs to be perturbation radius or perturbation"
+            " per feature:",
+            attack_model,
+        )
 
 
 @jit(nopython=True, nogil=NOGIL)
@@ -320,7 +451,7 @@ def _scan_numerical_feature_fast(
     return best_score, best_split
 
 
-# @jit(nopython=True, nogil=NOGIL)
+@jit(nopython=True, nogil=NOGIL)
 def _scan_numerical_feature_fast_regression(
     samples,
     y,
@@ -477,7 +608,7 @@ def adversarial_gini_gain_two_class(l_0, l_1, li_0, li_1, ri_0, ri_1, r_0, r_1):
     # Compute these before since we use their values multiple times
     x_coef = (l_0 + r_0 + i_0) / (l_1 + r_1 + i_1)
     intercept = (l_1 * r_0 - l_0 * r_1 - l_0 * i_1 + l_1 * i_0) / (l_1 + r_1 + i_1)
-    denominator = x_coef ** 2 + 1
+    denominator = x_coef**2 + 1
 
     # In the paper we refer to m1, m0 here they are li_1 and li_0
     x_prime = round((li_1 + x_coef * (li_0 - intercept)) / denominator)
@@ -564,7 +695,7 @@ def gini_impurity(i_0, i_1):
         return 1.0
 
     ratio = i_0 / (i_0 + i_1)
-    return 1.0 - (ratio ** 2) - ((1 - ratio) ** 2)
+    return 1.0 - (ratio**2) - ((1 - ratio) ** 2)
 
 
 @jit(nopython=True, nogil=NOGIL)
@@ -649,8 +780,8 @@ def weighted_gini(l_0, l_1, r_0, r_1):
     else:
         r_p = r_0 / (r_0 + r_1)
 
-    gini = l_t * (1 - (l_p ** 2) - ((1 - l_p) ** 2)) + r_t * (
-        1 - (r_p ** 2) - ((1 - r_p) ** 2)
+    gini = l_t * (1 - (l_p**2) - ((1 - l_p) ** 2)) + r_t * (
+        1 - (r_p**2) - ((1 - r_p) ** 2)
     )
 
     total = l_t + r_t
@@ -744,6 +875,7 @@ class BaseGrootTree(BaseEstimator):
         robust_weight=1.0,
         attack_model=None,
         chen_heuristic=False,
+        compile=True,
         random_state=None,
     ):
         self.max_depth = max_depth
@@ -753,6 +885,7 @@ class BaseGrootTree(BaseEstimator):
         self.robust_weight = robust_weight
         self.attack_model = attack_model
         self.chen_heuristic = chen_heuristic
+        self.compile = compile
         self.random_state = random_state
 
     def fit(self, X, y, check_input=True):
@@ -787,7 +920,9 @@ class BaseGrootTree(BaseEstimator):
 
         # Turn numerical features in attack model into tuples to make fitting
         # code simpler
-        self.attack_model_ = _attack_model_to_tuples(attack_model)
+        self.attack_model_ = np.array(
+            _attack_model_to_tuples(attack_model, X.shape[1]), dtype=X.dtype
+        )
 
         self.random_state_ = check_random_state(self.random_state)
 
@@ -810,6 +945,10 @@ class BaseGrootTree(BaseEstimator):
 
         self.root_ = self.__fit_recursive(X, y, constraints)
 
+        # Compile the tree into a representation that is faster when predicting
+        if self.compile:
+            self.compiled_root_ = CompiledTree(self.root_)
+
         return self
 
     def __fit_recursive(self, X, y, constraints, depth=0):
@@ -897,12 +1036,15 @@ class BaseGrootTree(BaseEstimator):
 
         return best_rule, best_feature, best_score
 
-    def to_string(self):
+    def to_string(self, feature_names=None):
         result = ""
         result += f"Parameters: {self.get_params()}\n"
 
+        if feature_names is None:
+            feature_names = [f"x[{i}]" for i in range(self.n_features_in_)]
+
         if hasattr(self, "root_"):
-            result += f"Tree:\n{self.root_.pretty_print()}"
+            result += f"Tree:\n{self.root_.pretty_print(feature_names)}"
         else:
             result += "Tree has not yet been fitted"
 
@@ -951,6 +1093,7 @@ class GrootTreeClassifier(BaseGrootTree, ClassifierMixin):
         attack_model=None,
         one_adversarial_class=False,
         chen_heuristic=False,
+        compile=True,
         random_state=None,
     ):
         """
@@ -972,6 +1115,8 @@ class GrootTreeClassifier(BaseGrootTree, ClassifierMixin):
             Whether one class (malicious, 1) perturbs their samples or if both classes (benign and malicious, 0 and 1) do so.
         chen_heuristic : bool, optional
             Whether to use the heuristic for the adversarial Gini impurity from Chen et al. (2019) instead of GROOT's adversarial Gini impurity.
+        compile : bool, optional
+            Whether to compile the tree for faster predictions.
         random_state : int, optional
             Controls the sampling of the features to consider when looking for the best split at each node.
 
@@ -987,6 +1132,8 @@ class GrootTreeClassifier(BaseGrootTree, ClassifierMixin):
             The number of features when `fit` is performed.
         root_ : Node
             The root node of the tree after fitting.
+        compiled_root_ : CompiledTree
+            The compiled root node of the tree after fitting.
         """
         self.max_depth = max_depth
         self.min_samples_split = min_samples_split
@@ -996,13 +1143,15 @@ class GrootTreeClassifier(BaseGrootTree, ClassifierMixin):
         self.attack_model = attack_model
         self.one_adversarial_class = one_adversarial_class
         self.chen_heuristic = chen_heuristic
+        self.compile = compile
         self.random_state = random_state
 
     def _check_target(self, y):
         target_type = type_of_target(y)
         if target_type != "binary":
             raise ValueError(
-                f"Unknown label type: classifier only supports binary labels but found {target_type}"
+                "Unknown label type: classifier only supports binary labels but found"
+                f" {target_type}"
             )
 
         self.classes_, y = np.unique(y, return_inverse=True)
@@ -1382,6 +1531,10 @@ class GrootTreeClassifier(BaseGrootTree, ClassifierMixin):
                 "Received different number of features during predict than during fit"
             )
 
+        # If model has been compiled, use compiled predict_proba
+        if self.compile:
+            return self.compiled_root_.predict_classification_proba(X)
+
         predictions = []
         for sample in X:
             predictions.append(self.root_.predict(sample))
@@ -1409,6 +1562,19 @@ class GrootTreeClassifier(BaseGrootTree, ClassifierMixin):
             The predicted class labels
         """
 
+        # If model has been compiled, use compiled predict
+        if self.compile:
+            check_is_fitted(self, "root_")
+
+            X = check_array(X)
+            if X.shape[1] != self.n_features_in_:
+                raise ValueError(
+                    "Received different number of features during predict than"
+                    " during fit"
+                )
+
+            return self.classes_.take(self.compiled_root_.predict_classification(X))
+
         y_pred_proba = self.predict_proba(X)
 
         return self.classes_.take(np.argmax(y_pred_proba, axis=1))
@@ -1430,6 +1596,7 @@ class GrootTreeRegressor(BaseGrootTree, RegressorMixin):
         robust_weight=1.0,
         attack_model=None,
         chen_heuristic=False,
+        compile=True,
         random_state=None,
     ):
         """
@@ -1451,6 +1618,8 @@ class GrootTreeRegressor(BaseGrootTree, RegressorMixin):
             Whether one class (malicious, 1) perturbs their samples or if both classes (benign and malicious, 0 and 1) do so.
         chen_heuristic : bool, optional
             Whether to use the heuristic for the adversarial Gini impurity from Chen et al. (2019) instead of GROOT's adversarial Gini impurity.
+        compile : bool, optional
+            Whether to compile the tree for faster predictions.
         random_state : int, optional
             Controls the sampling of the features to consider when looking for the best split at each node.
 
@@ -1466,6 +1635,8 @@ class GrootTreeRegressor(BaseGrootTree, RegressorMixin):
             The number of features when `fit` is performed.
         root_ : Node
             The root node of the tree after fitting.
+        compiled_root_ : CompiledTree
+            The compiled root node of the tree after fitting.
         """
         self.max_depth = max_depth
         self.min_samples_split = min_samples_split
@@ -1474,13 +1645,15 @@ class GrootTreeRegressor(BaseGrootTree, RegressorMixin):
         self.robust_weight = robust_weight
         self.attack_model = attack_model
         self.chen_heuristic = chen_heuristic
+        self.compile = compile
         self.random_state = random_state
 
     def _check_target(self, y):
         target_type = type_of_target(y)
         if target_type not in {"continuous", "multiclass", "binary"}:
             raise ValueError(
-                f"Unknown label type: regressor only supports continuous/multiclass/binary targets but found {target_type}"
+                "Unknown label type: regressor only supports"
+                f" continuous/multiclass/binary targets but found {target_type}"
             )
 
         # Make a copy of y if it is readonly to prevent errors
@@ -1626,6 +1799,10 @@ class GrootTreeRegressor(BaseGrootTree, RegressorMixin):
                 "Received different number of features during predict than during fit"
             )
 
+        # If model has been compiled, use compiled predict
+        if self.compile:
+            return self.compiled_root_.predict_regression(X)
+
         predictions = []
         for sample in X:
             predictions.append(self.root_.predict(sample))
@@ -1665,6 +1842,7 @@ class BaseGrootRandomForest(BaseEstimator):
         chen_heuristic=False,
         max_samples=None,
         n_jobs=None,
+        compile=True,
         random_state=None,
     ):
         """
@@ -1692,6 +1870,8 @@ class BaseGrootRandomForest(BaseEstimator):
             The fraction of samples to draw from X to train each decision tree. If None (default), then draw X.shape[0] samples.
         n_jobs : int, optional
             The number of jobs to run in parallel when fitting trees. See joblib.
+        compile : bool, optional
+            Whether to compile decision trees for faster predictions.
         random_state : int, optional
             Controls the sampling of the features to consider when looking for the best split at each node.
 
@@ -1715,6 +1895,7 @@ class BaseGrootRandomForest(BaseEstimator):
         self.chen_heuristic = chen_heuristic
         self.max_samples = max_samples
         self.n_jobs = n_jobs
+        self.compile = compile
         self.random_state = random_state
 
     def fit(self, X, y):
@@ -1826,6 +2007,7 @@ class GrootRandomForestClassifier(BaseGrootRandomForest, ClassifierMixin):
         chen_heuristic=False,
         max_samples=None,
         n_jobs=None,
+        compile=True,
         random_state=None,
     ):
         """
@@ -1855,6 +2037,8 @@ class GrootRandomForestClassifier(BaseGrootRandomForest, ClassifierMixin):
             The fraction of samples to draw from X to train each decision tree. If None (default), then draw X.shape[0] samples.
         n_jobs : int, optional
             The number of jobs to run in parallel when fitting trees. See joblib.
+        compile : bool, optional
+            Whether to compile decision trees for faster predictions.
         random_state : int, optional
             Controls the sampling of the features to consider when looking for the best split at each node.
 
@@ -1879,6 +2063,7 @@ class GrootRandomForestClassifier(BaseGrootRandomForest, ClassifierMixin):
             chen_heuristic=chen_heuristic,
             max_samples=max_samples,
             n_jobs=n_jobs,
+            compile=compile,
             random_state=random_state,
         )
         self.one_adversarial_class = one_adversarial_class
@@ -1887,7 +2072,8 @@ class GrootRandomForestClassifier(BaseGrootRandomForest, ClassifierMixin):
         target_type = type_of_target(y)
         if target_type != "binary":
             raise ValueError(
-                f"Unknown label type: classifier only supports binary labels but found {target_type}"
+                "Unknown label type: classifier only supports binary labels but found"
+                f" {target_type}"
             )
 
         self.classes_, y = np.unique(y, return_inverse=True)
@@ -1905,6 +2091,7 @@ class GrootRandomForestClassifier(BaseGrootRandomForest, ClassifierMixin):
             attack_model=self.attack_model,
             one_adversarial_class=self.one_adversarial_class,
             chen_heuristic=self.chen_heuristic,
+            compile=self.compile,
         )
 
     def predict_proba(self, X):
@@ -1976,6 +2163,7 @@ class GrootRandomForestRegressor(BaseGrootRandomForest, RegressorMixin):
         chen_heuristic=False,
         max_samples=None,
         n_jobs=None,
+        compile=True,
         random_state=None,
     ):
         """
@@ -2003,6 +2191,8 @@ class GrootRandomForestRegressor(BaseGrootRandomForest, RegressorMixin):
             The fraction of samples to draw from X to train each decision tree. If None (default), then draw X.shape[0] samples.
         n_jobs : int, optional
             The number of jobs to run in parallel when fitting trees. See joblib.
+        compile : bool, optional
+            Whether to compile decision trees for faster predictions.
         random_state : int, optional
             Controls the sampling of the features to consider when looking for the best split at each node.
 
@@ -2027,6 +2217,7 @@ class GrootRandomForestRegressor(BaseGrootRandomForest, RegressorMixin):
             chen_heuristic=chen_heuristic,
             max_samples=max_samples,
             n_jobs=n_jobs,
+            compile=compile,
             random_state=random_state,
         )
 
@@ -2034,7 +2225,8 @@ class GrootRandomForestRegressor(BaseGrootRandomForest, RegressorMixin):
         target_type = type_of_target(y)
         if target_type not in {"continuous", "multiclass", "binary"}:
             raise ValueError(
-                f"Unknown label type: regressor only supports continuous/multiclass/binary targets but found {target_type}"
+                "Unknown label type: regressor only supports"
+                f" continuous/multiclass/binary targets but found {target_type}"
             )
 
         # Make a copy of y if it is readonly to prevent errors
@@ -2052,6 +2244,7 @@ class GrootRandomForestRegressor(BaseGrootRandomForest, RegressorMixin):
             robust_weight=self.robust_weight,
             attack_model=self.attack_model,
             chen_heuristic=self.chen_heuristic,
+            compile=self.compile,
         )
 
     def predict(self, X):
